@@ -21,12 +21,12 @@
 
 #include "config.h"
 
-#include "galias.h"
-
 #include "goption.h"
 #include "glib.h"
 #include "gi18n.h"
 
+#include "galias.h"
+
 #include <string.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -312,6 +312,13 @@ g_option_context_set_main_group (GOptionContext *context,
   g_return_if_fail (context != NULL);
   g_return_if_fail (group != NULL);
 
+  if (context->main_group)
+    {
+      g_warning ("This GOptionContext already has a main group");
+
+      return;
+    }
+
   context->main_group = group;
 }
 
@@ -362,6 +369,35 @@ g_option_context_add_main_entries (GOptionContext      *context,
   g_option_group_set_translation_domain (context->main_group, translation_domain);
 }
 
+static gint
+calculate_max_length (GOptionGroup *group)
+{
+  GOptionEntry *entry;
+  gint i, len, max_length;
+
+  max_length = 0;
+
+  for (i = 0; i < group->n_entries; i++)
+    {
+      entry = &group->entries[i];
+
+      if (entry->flags & G_OPTION_FLAG_HIDDEN)
+	continue;
+
+      len = g_utf8_strlen (entry->long_name, -1);
+      
+      if (entry->short_name)
+	len += 4;
+      
+      if (entry->arg != G_OPTION_ARG_NONE && entry->arg_description)
+	len += 1 + g_utf8_strlen (TRANSLATE (group, entry->arg_description), -1);
+      
+      max_length = MAX (max_length, len);
+    }
+
+  return max_length;
+}
+
 static void
 print_entry (GOptionGroup       *group,
 	     gint                max_length,
@@ -372,6 +408,9 @@ print_entry (GOptionGroup       *group,
   if (entry->flags & G_OPTION_FLAG_HIDDEN)
     return;
 
+  if (entry->long_name[0] == 0)
+    return;
+
   str = g_string_new (NULL);
   
   if (entry->short_name)
@@ -390,11 +429,12 @@ print_entry (GOptionGroup       *group,
 static void
 print_help (GOptionContext *context,
 	    gboolean        main_help,
-	    GOptionGroup   *group)
+	    GOptionGroup   *group) 
 {
   GList *list;
   gint max_length, len;
   gint i;
+  GOptionEntry *entry;
   GHashTable *shadow_map;
   gboolean seen[256];
   
@@ -409,14 +449,15 @@ print_help (GOptionContext *context,
     {
       for (i = 0; i < context->main_group->n_entries; i++)
 	{
+	  entry = &context->main_group->entries[i];
 	  g_hash_table_insert (shadow_map, 
-			       (gpointer)context->main_group->entries[i].long_name, 
-			       context->main_group->entries + i);
+			       (gpointer)entry->long_name, 
+			       entry);
 	  
-	  if (seen[(guchar)context->main_group->entries[i].short_name])
-	    context->main_group->entries[i].short_name = 0;
+	  if (seen[(guchar)entry->short_name])
+	    entry->short_name = 0;
 	  else
-	    seen[(guchar)context->main_group->entries[i].short_name] = TRUE;
+	    seen[(guchar)entry->short_name] = TRUE;
 	}
     }
 
@@ -426,15 +467,16 @@ print_help (GOptionContext *context,
       GOptionGroup *group = list->data;
       for (i = 0; i < group->n_entries; i++)
 	{
-	  if (g_hash_table_lookup (shadow_map, group->entries[i].long_name))
-	    group->entries[i].long_name = g_strdup_printf ("%s-%s", group->name, group->entries[i].long_name);
+	  entry = &group->entries[i];
+	  if (g_hash_table_lookup (shadow_map, entry->long_name))
+	    entry->long_name = g_strdup_printf ("%s-%s", group->name, entry->long_name);
 	  else  
-	    g_hash_table_insert (shadow_map, (gpointer)group->entries[i].long_name, group->entries + i);
+	    g_hash_table_insert (shadow_map, (gpointer)entry->long_name, entry);
 
-	  if (seen[(guchar)group->entries[i].short_name])
-	    group->entries[i].short_name = 0;
+	  if (seen[(guchar)entry->short_name])
+	    entry->short_name = 0;
 	  else
-	    seen[(guchar)group->entries[i].short_name] = TRUE;
+	    seen[(guchar)entry->short_name] = TRUE;
 	}
       list = list->next;
     }
@@ -443,7 +485,7 @@ print_help (GOptionContext *context,
 
   list = context->groups;
 
-  max_length = g_utf8_strlen ("--help, -?", -1);
+  max_length = g_utf8_strlen ("-?, --help", -1);
 
   if (list)
     {
@@ -451,6 +493,12 @@ print_help (GOptionContext *context,
       max_length = MAX (max_length, len);
     }
 
+  if (context->main_group)
+    {
+      len = calculate_max_length (context->main_group);
+      max_length = MAX (max_length, len);
+    }
+
   while (list != NULL)
     {
       GOptionGroup *group = list->data;
@@ -460,48 +508,40 @@ print_help (GOptionContext *context,
       max_length = MAX (max_length, len);
 
       /* Then we go through the entries */
-      for (i = 0; i < group->n_entries; i++)
-	{
-	  if (group->entries[i].flags & G_OPTION_FLAG_HIDDEN)
-	    continue;
-
-	  len = g_utf8_strlen (group->entries[i].long_name, -1);
-
-	  if (group->entries[i].short_name)
-	    len += 4;
-
-	  if (group->entries[i].arg != G_OPTION_ARG_NONE &&
-	      group->entries[i].arg_description)
-	    len += 1 + g_utf8_strlen (TRANSLATE (group, group->entries[i].arg_description), -1);
-
-	  max_length = MAX (max_length, len);
-	}
+      len = calculate_max_length (group);
+      max_length = MAX (max_length, len);
       
       list = list->next;
     }
 
   /* Add a bit of padding */
   max_length += 4;
-  
-  list = context->groups;
 
-  g_print ("%s\n  --%-*s %s\n", 
-	   _("Help Options:"), max_length, "help", _("Show help options"));
-
-  /* We only want --help-all when there are groups */
-  if (list)
-    g_print ("  --%-*s %s\n", max_length, "help-all", _("Show all help options"));
-
-  while (list)
+  if (!group)
     {
-      GOptionGroup *group = list->data;
-
-      g_print ("  --help-%-*s %s\n", max_length - 5, group->name, TRANSLATE (group, group->help_description));
+      list = context->groups;
       
-      list = list->next;
-    }
+      g_print ("%s\n  -%c, --%-*s %s\n", 
+	       _("Help Options:"), '?', max_length - 4, "help", 
+	       _("Show help options"));
+      
+      /* We only want --help-all when there are groups */
+      if (list)
+	g_print ("  --%-*s %s\n", max_length, "help-all", 
+		 _("Show all help options"));
+      
+      while (list)
+	{
+	  GOptionGroup *group = list->data;
+	  
+	  g_print ("  --help-%-*s %s\n", max_length - 5, group->name, 
+		   TRANSLATE (group, group->help_description));
+	  
+	  list = list->next;
+	}
 
-  g_print ("\n");
+      g_print ("\n");
+    }
 
   if (group)
     {
@@ -542,7 +582,8 @@ print_help (GOptionContext *context,
 
       if (context->main_group)
 	for (i = 0; i < context->main_group->n_entries; i++) 
-	  print_entry (context->main_group, max_length, &context->main_group->entries[i]);
+	  print_entry (context->main_group, max_length, 
+		       &context->main_group->entries[i]);
 
       while (list != NULL)
 	{
@@ -569,10 +610,11 @@ parse_int (const gchar *arg_name,
 	   GError     **error)
 {
   gchar *end;
-  glong tmp = strtol (arg, &end, 0);
+  glong tmp; 
 
   errno = 0;
-  
+  tmp = strtol (arg, &end, 0);
+
   if (*arg == '\0' || *end != '\0')
     {
       g_set_error (error,
@@ -1054,8 +1096,10 @@ free_pending_nulls (GOptionContext *context,
  *
  * If the parsing is successful, any parsed arguments are
  * removed from the array and @argc and @argv are updated 
- * accordingly. In case of an error, @argc and @argv are
- * left unmodified.
+ * accordingly. A '--' option is stripped from @argv
+ * unless there are unparsed options before and after it, 
+ * or some of the options after it start with '-'. In case 
+ * of an error, @argc and @argv are left unmodified. 
  *
  * If automatic <option>--help</option> support is enabled
  * (see g_option_context_set_help_enabled()), and the 
@@ -1117,13 +1161,15 @@ g_option_context_parse (GOptionContext   *context,
   if (argc && argv)
     {
       gboolean stop_parsing = FALSE;
+      gboolean has_unknown = FALSE;
+      gint separator_pos = 0;
 
       for (i = 1; i < *argc; i++)
 	{
 	  gchar *arg, *dash;
 	  gboolean parsed = FALSE;
 
-	  if ((*argv)[i][0] == '-' && !stop_parsing)
+	  if ((*argv)[i][0] == '-' && (*argv)[i][1] != '\0' && !stop_parsing)
 	    {
 	      if ((*argv)[i][1] == '-')
 		{
@@ -1134,7 +1180,7 @@ g_option_context_parse (GOptionContext   *context,
 		  /* '--' terminates list of arguments */
 		  if (*arg == 0)
 		    {
-		      add_pending_null (context, &((*argv)[i]), NULL);
+		      separator_pos = i;
 		      stop_parsing = TRUE;
 		      continue;
 		    }
@@ -1234,6 +1280,9 @@ g_option_context_parse (GOptionContext   *context,
 		  
 		  for (j = 0; j < strlen (arg); j++)
 		    {
+		      if (context->help_enabled && arg[j] == '?')
+			print_help (context, TRUE, NULL);
+		      
 		      parsed = FALSE;
 		      
 		      if (context->main_group &&
@@ -1287,7 +1336,7 @@ g_option_context_parse (GOptionContext   *context,
 			  if (!nulled_out[j])
 			    {
 			      if (!new_arg)
-				new_arg = g_malloc (strlen (arg));
+				new_arg = g_malloc (strlen (arg) + 1);
 			      new_arg[arg_index++] = arg[j];
 			    }
 			}
@@ -1303,11 +1352,14 @@ g_option_context_parse (GOptionContext   *context,
 		    }
 		}
 	      
+	      if (!parsed)
+		has_unknown = TRUE;
+
 	      if (!parsed && !context->ignore_unknown)
 		{
 		  g_set_error (error,
 			       G_OPTION_ERROR, G_OPTION_ERROR_UNKNOWN_OPTION,
-			       _("Unknown option %s"), (*argv)[i]);
+				   _("Unknown option %s"), (*argv)[i]);
 		  goto fail;
 		}
 	    }
@@ -1319,32 +1371,41 @@ g_option_context_parse (GOptionContext   *context,
 					argc, argv, error, &parsed))
 		goto fail;
 	      
+	      if (!parsed && (has_unknown || (*argv)[i][0] == '-'))
+		separator_pos = 0;
 	    }
 	}
 
-      /* Call post-parse hooks */
-      list = context->groups;
-      while (list)
-	{
-	  GOptionGroup *group = list->data;
+      if (separator_pos > 0)
+	add_pending_null (context, &((*argv)[separator_pos]), NULL);
+	
+    }
 
-	  if (group->post_parse_func)
-	    {
-	      if (!(* group->post_parse_func) (context, group,
-					       group->user_data, error))
-		goto fail;
-	    }
-	  
-	  list = list->next;
-	}
-
-      if (context->main_group && context->main_group->post_parse_func)
+  /* Call post-parse hooks */
+  list = context->groups;
+  while (list)
+    {
+      GOptionGroup *group = list->data;
+      
+      if (group->post_parse_func)
 	{
-	  if (!(* context->main_group->post_parse_func) (context, context->main_group,
-							 context->main_group->user_data, error))
+	  if (!(* group->post_parse_func) (context, group,
+					   group->user_data, error))
 	    goto fail;
 	}
-
+      
+      list = list->next;
+    }
+  
+  if (context->main_group && context->main_group->post_parse_func)
+    {
+      if (!(* context->main_group->post_parse_func) (context, context->main_group,
+						     context->main_group->user_data, error))
+	goto fail;
+    }
+  
+  if (argc && argv)
+    {
       free_pending_nulls (context, TRUE);
       
       for (i = 1; i < *argc; i++)
@@ -1604,3 +1665,5 @@ g_option_group_set_translation_domain (GOptionGroup *group,
 				     g_free);
 } 
 
+#define __G_OPTION_C__
+#include "galiasdef.c"
